/* Crypton Client, Copyright 2013 SpiderOak, Inc.
 *
 * This file is part of Crypton Client.
 *
 * Crypton Client is free software: you can redistribute it and/or modify it
 * under the terms of the Affero GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * Crypton Client is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the Affero GNU General Public
 * License for more details.
 *
 * You should have received a copy of the Affero GNU General Public License
 * along with Crypton Client.  If not, see <http://www.gnu.org/licenses/>.
*/

var assert = chai.assert;

describe('Container', function () {
  describe('add()', function () {
    var container = new crypton.Container();

    it('should add a key to the keys array', function (done) {
      container.add('legit', function (err) {
        assert.equal(err, null);
        assert.deepEqual(container.keys.legit, {});
        done();
      });
    });

    it('should refuse to add an existing key', function (done) {
      container.add('legit', function (err) {
        assert.equal(err, 'Key already exists');
        done();
      });
    });
  });

  describe('get()', function () {
    var container = new crypton.Container();

    before(function (done) {
      container.add('legit', function () {
        done();
      });
    });

    it('should return the correct key', function (done) {
      container.get('legit', function (err, legit) {
        assert.equal(err, null);
        assert.deepEqual(container.keys.legit, legit);
        done();
      });
    });

    it('should refuse to return a nonexistent key', function (done) {
      container.get('grail', function (err) {
        assert.equal(err, 'Key does not exist');
        done();
      });
    });
  });

  describe('save()', function () {
    it('should err if the container hasn\'t changed', function (done) {
      var container = new crypton.Container();
      container.save(function (err) {
        assert.equal(err, 'Container has not changed');
        done();
      });
    });

    it('should add a new version', function (done) {
      var container = new crypton.Container();
      var now = container.version;

      // these are necessary for the encryption
      // typically you will allow the session to create
      // a container instead of doing it manually
      container.name = 'legit';
      container.session = {};
      container.session.account = {};
      container.session.account.containerNameHmacKey = [1193696192,274367050,-1647541843,-73767300,1167252974,-984408945,-1161509559,962393744];
      container.hmacKey = [644454176,567210777,1585131513,-1319788074,1059693219,-1045618495,1515382626,2119063568];

      container.add('foo', function () {
        container.get('foo', function (err, foo) {
          foo.bar = 'baz';
          container.save(function (err) {
            assert.notEqual(container.version > now);
            var newVersion = container.versions[container.version];
            assert.deepEqual(newVersion.foo, { bar: 'baz' });
            done();
          }, {
            save: false
          });
        });
      });
    });

    it('should generate the correct transaction chunk', function (done) {
      var container = new crypton.Container();

      // these are necessary for the encryption
      // typically you will allow the session to create
      // a container instead of doing it manually
      container.name = 'legit';
      container.session = {};
      container.session.account = {};
      container.session.account.containerNameHmacKey = [1193696192,274367050,-1647541843,-73767300,1167252974,-984408945,-1161509559,962393744];
      container.hmacKey = [644454176,567210777,1585131513,-1319788074,1059693219,-1045618495,1515382626,2119063568];

      container.add('foo', function () {
        container.get('foo', function (err, foo) {
          foo.bar = 'baz';
          container.save(function (err, chunk) {
            assert.equal(err, null);
            assert.equal(chunk.type, 'addContainerRecord');
            assert.equal(chunk.containerNameHmac, 'a6b4fdf950fb02d66bee39fbe8af8456a615969275a2ec0db0fc09936b99e882');
            assert.deepEqual(Object.keys(JSON.parse(chunk.payloadCiphertext)), [
              'iv',
              'v',
              'iter',
              'ks',
              'ts',
              'mode',
              'adata',
              'cipher',
              'ct'
            ]);
            done();
          }, {
            save: false
          });
        });
      });
    });
  });

  describe('getDiff()', function () {
    // TODO this would be a good place to
    // test more complex interactions!
    it('should get the correct diff', function (done) {
      var container = new crypton.Container();
      container.add('foo', function () {
        container.get('foo', function (err, foo) {
          foo.bar = 'baz';
          container.getDiff(function (err, diff) {
            assert.equal(err, null);
            assert.deepEqual(diff, { foo: [ { bar: 'baz' } ] });
            done();
          });
        });
      });
    });

    it('should callback with null on an unchanged container', function (done) {
      var container = new crypton.Container();
      container.getDiff(function (err, diff) {
        assert.equal(err, null);
        assert.equal(diff, null);
        done();
      });
    });
  });

  describe('getVersions()', function () {
    it('should return an array', function () {
      var container = new crypton.Container();
      var ret = container.getVersions();
      assert.isArray(ret);
    });

    it('should return the correct array', function () {
      var container = new crypton.Container();
      var now = +new Date();
      container.versions[now] = {};
      var ret = container.getVersions();
      assert.deepEqual(ret, [now+'']);
      // version keys will always be strings
    });
  });

  describe('getVersion()', function () {
    it('should return the correct object', function () {
      var container = new crypton.Container();
      var now = +new Date();
      var expected = { foo: 'bar' };
      container.versions[now] = expected;
      var ret = container.getVersion(now);
      assert.deepEqual(ret, expected);
    });
  });

  describe('latestVersion()', function () {
    it('should return the latest version timestamp', function () {
      var container = new crypton.Container();
      var last;

      for (var i = 0; i < 1000; i++) {
        last = +new Date();
        container.versions[last] = {};
      }

      assert.equal(container.latestVersion(), last);
    });
  });

  describe('getPublicName()', function () {
    it('should return the correct hmac for a standard name', function () {
      var container = new crypton.Container();

      // these are necessary for the encryption
      // typically you will allow the session to create
      // a container instead of doing it manually
      container.name = 'legit';
      container.session = {};
      container.session.account = {};
      container.session.account.containerNameHmacKey = [1193696192,274367050,-1647541843,-73767300,1167252974,-984408945,-1161509559,962393744];
      container.hmacKey = [644454176,567210777,1585131513,-1319788074,1059693219,-1045618495,1515382626,2119063568];

      var expected = 'a6b4fdf950fb02d66bee39fbe8af8456a615969275a2ec0db0fc09936b99e882';
      assert.equal(container.getPublicName(), expected);
    });
  });

  describe('parseHistory()', function () {
    // TODO this should go in the integration tests
    // because it requires a valid session
  });

  describe('decryptRecord()', function () {
    // TODO this should go in the integration tests
    // because it requires a valid session
  });

  describe('getHistory()', function () {
    // TODO this should go in the integration tests
    // because it requires a valid session
  });

  describe('sync()', function () {
    // TODO this should go in the integration tests
    // because it requires a valid session
  });
});
